use std::collections::HashMap;
use serde::{Deserialize, Serialize};
use actix_web::{web, HttpRequest, HttpResponse, Responder};
use chrono::NaiveDateTime;
use crate::models::models::Comment;
use crate::repositories::comment::{create_comment, get_issue_comments};
use crate::repositories::session::get_username_from_session;

#[derive(Deserialize)]
pub struct AddCommentRequest {
    pub content: String,
    pub parent: Option<i64>,
    pub issue_id: i32,
}

#[derive(Serialize)]
pub struct AddCommentResponse {
    pub content: String,
    pub parent: Option<i64>,
    pub telegram_handle: String,
    pub created_at: NaiveDateTime,
}

pub async fn add_comment(req: HttpRequest, data: web::Json<AddCommentRequest>) -> impl Responder {
    if let Some(session) = req.cookie("session") {
        if let Some(username) = get_username_from_session(session.value().to_string()) {
            let new_comment = create_comment(data.content.clone(), data.parent, username.clone(), data.issue_id);
            HttpResponse::Ok().json(
                AddCommentResponse {
                    content: new_comment.content,
                    parent: new_comment.parent,
                    telegram_handle: username.clone(),
                    created_at: new_comment.created_at.unwrap()
                })
        } else {
            HttpResponse::Unauthorized().into()
        }
    } else {
        HttpResponse::Unauthorized().into()
    }
}

#[derive(Deserialize)]
pub struct GetCommentsRequest {
    pub issue_id: i32,
}

#[derive(Serialize, Clone)]
pub struct CommentWithChildren {
    pub comment: Comment,
    pub children: Vec<CommentWithChildren>,
}

#[derive(Serialize)]
pub struct GetCommentsResponse {
    pub comments: Vec<CommentWithChildren>,
}

fn build_comment_tree(comments: Vec<Comment>) -> Vec<CommentWithChildren> {
    let mut comment_map: HashMap<i64, CommentWithChildren> = HashMap::new();
    let mut child_map: HashMap<i64, Vec<i64>> = HashMap::new();
    let mut root_ids: Vec<i64> = Vec::new();

    for comment in comments {
        let id = comment.id;
        let comment_with_children = CommentWithChildren {
            comment: comment.clone(),
            children: Vec::new(),
        };
        comment_map.insert(id, comment_with_children);

        if let Some(parent_id) = comment.parent {
            child_map.entry(parent_id).or_default().push(id);
        } else {
            root_ids.push(id);
        }
    }

    fn build_tree(id: i64, comment_map: &mut HashMap<i64, CommentWithChildren>, child_map: &HashMap<i64, Vec<i64>>) -> CommentWithChildren {
        let mut comment = comment_map.remove(&id).unwrap();
        if let Some(child_ids) = child_map.get(&id) {
            comment.children = child_ids.iter()
                .map(|&child_id| build_tree(child_id, comment_map, child_map))
                .collect();
        }
        comment
    }

    root_ids.into_iter()
        .map(|id| build_tree(id, &mut comment_map, &child_map))
        .collect()
}

pub async fn get_comments(data: web::Query<GetCommentsRequest>) -> impl Responder {
    let comments = build_comment_tree(get_issue_comments(data.issue_id));

    HttpResponse::Ok().json(GetCommentsResponse {
        comments
    })
}