use crate::db::db::establish_connection;
use crate::models::models::{Issue, IssueWithSummaryAndVotes, NewIssue, NewIssueVote, NewParagraph, Paragraph};
use crate::schema::issue_votes::{issue_id, positive};
use crate::schema::paragraphs;
use crate::schema::paragraphs::{content, index};
use crate::schema::{issue_votes, issues};
use chrono::NaiveDateTime;
use diesel;
use diesel::associations::HasTable;
use diesel::dsl::count;
use diesel::result::DatabaseErrorKind;
use diesel::result::Error::DatabaseError;
use diesel::{sql_query, BoolExpressionMethods, ExpressionMethods, QueryDsl, RunQueryDsl, SelectableHelper};
use serde::Deserialize;

#[derive(Deserialize)]
pub struct FilterOptions {
    pub min_positive_votes: Option<i64>,
    pub min_votes: Option<i64>,
}

pub fn get_issues(filters: Option<FilterOptions>, offset: i32, limit: i16) -> Vec<IssueWithSummaryAndVotes> {
    let connection = &mut establish_connection();

    let query = sql_query("select i.id, title, content as summary, i.paragraph_count, telegram_handle, i.created_at,
        count(v.id) as total_votes,
        sum(case when v.positive then 1 else 0 end) as positive_votes
    from issues i
        inner join paragraphs p on i.id = p.post_id
        inner join issue_votes v on i.id = v.issue_id
    where p.index = 0
    group by v.issue_id, title, telegram_handle, content, i.id;");

    let mut issues_with_content: Vec<IssueWithSummaryAndVotes> = query
        .load::<IssueWithSummaryAndVotes>(connection)
        .unwrap_or_default()
        .into_iter()
        .collect();

    if let Some(filter_info) = filters {
        if let Some(positive_votes) = filter_info.min_positive_votes {
            issues_with_content = issues_with_content
                .iter()
                .filter(|i| i.positive_votes >= positive_votes)
                .map(|i| i.clone())
                .collect()
        }
        if let Some(min_votes) = filter_info.min_votes {
            issues_with_content = issues_with_content
                .iter()
                .filter(|i| i.total_votes >= min_votes)
                .map(|i| i.clone())
                .collect()
        }
    }

    let len = issues_with_content.len();
    let i1 = offset.max(0) as usize;
    let i2 = i1.saturating_add(limit as usize).min(len);
    issues_with_content[i1..i2].to_vec()
}

pub fn create_issue(title: &String, paragraphs: &Vec<String>, telegram_handle: &Option<String>) -> Option<Issue> {
    let connection = &mut establish_connection();

    if let Some(tg_handle) = telegram_handle {
        let new_issue = NewIssue {
            title: title.to_string(),
            paragraph_count: paragraphs.len() as i16,
            telegram_handle: tg_handle.to_string(),
        };
        let insertion = diesel::insert_into(issues::table)
            .values(&new_issue)
            .returning(Issue::as_select())
            .get_result(connection)
            .expect("Error saving new issue");

        let _ = diesel::insert_into(paragraphs::table)
            .values(paragraphs.iter().enumerate().map(|(i, p)|
                NewParagraph {
                    content: Some(p.clone()),
                    index: i as i32,
                    post_id: insertion.id
                }).collect::<Vec<NewParagraph>>())
            .returning(Paragraph::as_select())
            .get_results(connection)
            .expect("Error saving paragraphs");

        return Some(insertion)
    }
    None
}

pub fn get_paragraphs(other_issue_id: i32) -> Vec<Option<String>> {
    let connection = &mut establish_connection();
    paragraphs::table
        .filter(paragraphs::post_id.eq(other_issue_id))
        .order(index)
        .select(content)
        .load::<String>(connection)
        .unwrap_or_default()
        .into_iter()
        .map(Some)
        .collect::<Vec<Option<String>>>()
}


pub fn record_vote_for_issue(other_positive: bool, other_issue_id: i32, user_id: i64) -> bool {
    let connection = &mut establish_connection();

    let result = diesel::insert_into(issue_votes::table)
        .values(NewIssueVote { positive: other_positive, issue_id: other_issue_id, user_id })
        .execute(connection);

    match result {
        Ok(_) => true,
        Err(DatabaseError(kind, _)) => {
            if let DatabaseErrorKind::UniqueViolation = kind {
                let _ = diesel::update(
                    issue_votes::table
                        .filter(
                            issue_votes::user_id.eq(user_id).and(issue_votes::issue_id.eq(issue_id))
                        )
                )
                    .set(issue_votes::positive.eq(positive))
                    .execute(connection);
                return true
            }
            false
        },
        _ => false
    }
}

pub fn get_issue_vote_equity(issue_identifier: i32) -> i64 {
    let connection = &mut establish_connection();
    let positive_votes: i64 = issue_votes::table
        .filter(issue_votes::issue_id.eq(issue_identifier).and(issue_votes::positive.eq(true)))
        .count()
        .get_result(connection)
        .unwrap();
    let negative_votes: i64 = issue_votes::table
        .filter(issue_votes::issue_id.eq(issue_identifier).and(issue_votes::positive.eq(false)))
        .count()
        .get_result(connection)
        .unwrap();

    positive_votes - negative_votes
}

pub fn get_vote_for_user(other_issue_id: i32, other_user_id: i64) -> Option<bool> {
    let connection = &mut establish_connection();
    let result: Option<bool> = issue_votes::table
        .filter(issue_votes::issue_id.eq(other_issue_id).and(issue_votes::user_id.eq(other_user_id)))
        .limit(1)
        .select(issue_votes::positive)
        .get_result::<Option<bool>>(connection)
        .unwrap_or(None);

    result
}
